/*
 * kgraft_patch_bsc1025254
 *
 * Fix for bsc#1025254
 *
 *  SLE12-SP2 commit:
 *  f74d718d638e7012bbf32c43c2b4b88021e4ed26
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>

#include "local.h"

/* inlined */
static bool kgr_need_to_wait_for_sync(struct r1conf *conf, struct bio *bio)
{
	bool wait = false;

	if (test_bit(R1FLAG_ARRAY_FROZEN, &conf->flags) || !bio) {
		atomic_inc(&conf->nr_waiting);
		wait = true;
	}
	else if (atomic_read(&conf->barrier) && bio_data_dir(bio) == WRITE) {
		spin_lock_irq(&conf->resync_lock);
		if ((conf->mddev->curr_resync_completed
		     >= bio_end_sector(bio)) ||
		    (conf->next_resync + NEXT_NORMALIO_DISTANCE
		     <= bio->bi_iter.bi_sector))
			wait = false;
		else {
			atomic_inc(&conf->nr_waiting);
			wait = true;
		}
		spin_unlock_irq(&conf->resync_lock);
	}

	return wait;
}

/* patched function */
sector_t kgr_wait_barrier(struct r1conf *conf, struct bio *bio)
{
	sector_t sector = 0;

	if (kgr_need_to_wait_for_sync(conf, bio)) {
		spin_lock_irq(&conf->resync_lock);
		/* Wait for the barrier to drop.
		 * However if there are already pending
		 * requests (preventing the barrier from
		 * rising completely), and the
		 * per-process bio queue isn't empty,
		 * then don't wait, as we need to empty
		 * that queue to allow conf->start_next_window
		 * to increase.
		 */
		wait_event_lock_irq(conf->wait_barrier,
				    !test_bit(R1FLAG_ARRAY_FROZEN, &conf->flags) &&
				    (!atomic_read(&conf->barrier) ||
				     ((conf->start_next_window <
				       conf->next_resync + RESYNC_SECTORS) &&
				      current->bio_list &&
				      !bio_list_empty(current->bio_list))),
				    conf->resync_lock);
		/* nr_waiting got incremented in need_to_wait_for_sync() */
		atomic_dec(&conf->nr_waiting);
		spin_unlock_irq(&conf->resync_lock);
	}

	/*
	 * Fix bsc#1025254
	 *  -if (bio && bio_data_dir(bio) == WRITE &&
	 *  -    atomic_read(&conf->barrier)) {
	 *  +1 line
	 */
	if (bio && bio_data_dir(bio) == WRITE) {
		spin_lock_irq(&conf->resync_lock);
		if (bio->bi_iter.bi_sector >= conf->next_resync) {
			if (conf->start_next_window == MaxSector)
				conf->start_next_window =
					conf->next_resync +
					NEXT_NORMALIO_DISTANCE;

			if ((conf->start_next_window + NEXT_NORMALIO_DISTANCE)
			    <= bio->bi_iter.bi_sector)
				conf->next_window_requests++;
			else
				conf->current_window_requests++;
			sector = conf->start_next_window;
		}
		spin_unlock_irq(&conf->resync_lock);
	}
	atomic_inc(&conf->nr_pending);
	return sector;
}
